#ifndef MATRIX_HPP
#define MATRIX_HPP

#include <cassert>
#include <cstring>
#include <iostream>
#include <initializer_list>
#include "rational.hpp"

using namespace std;

typedef unsigned short int usint;

class Matrix{
private:
  class Line;
  class ConstLine;
  usint p,q;
  Rational* coeffs;
public:
  Matrix(usint p,usint q); 
  Matrix(const Matrix& M);
  ~Matrix();
  Matrix& operator=(const Matrix& M);
  usint lines() const;
  usint columns() const;
  Line operator[](usint l); 
  ConstLine operator[](usint l) const; 
  void set(initializer_list<Rational> list);
  void set(usint i,usint j,const Rational& x);
  Rational get(usint i,usint j) const;
  Rational& getRef(usint i,usint j);
  bool operator==(const Matrix& M) const;
  bool operator!=(const Matrix& M) const;
  Matrix operator+(const Matrix& M) const;
  Matrix operator-(const Matrix& M) const;
  Matrix operator*(const Rational& lambda) const;
  Matrix operator-() const;
  friend Matrix operator*(const Rational& lambda,const Matrix& M); 
  Matrix operator*(const Matrix& M) const;
  void swapLines(usint i,usint j);  //Li <-> Lj 
  void mulLine(usint i,Rational c); //Li <- c Li
  void addMulLine(usint i,usint j, Rational c); //Li <- Li + c Lj
  int findPivot(usint i,usint j);
  void Gauss();
  Rational determinant() const;
  bool isInvertible() const;
  Matrix inverse() const;
  int rank() const;
  friend ostream& operator<<(ostream& os,const Matrix& M);
};

class Matrix::Line{
 public:
  Matrix* matrix;
  usint l;
  Line(Matrix* M,usint l);
  Rational& operator[](usint c);
};

class Matrix::ConstLine{
 public:
  const Matrix *matrix;
  usint l;
  ConstLine(const Matrix *M,usint l);
  Rational operator[](usint c) const;
};

inline 
Matrix::Line Matrix::operator[](usint l){
  return Matrix::Line(this,l);
}

inline 
Matrix::ConstLine Matrix::operator[](usint l) const{
  return Matrix::ConstLine(this,l);
}

inline 
Matrix::Line::Line(Matrix* M,usint _l){
  matrix=M;
  l=_l;
}

inline Rational&
Matrix::Line::operator[](usint c){
  return matrix->getRef(l,c);
}

inline 
Matrix::ConstLine::ConstLine(const Matrix* M,usint _l){
  matrix=M;
  l=_l;
}

inline Rational
Matrix::ConstLine::operator[](usint c) const{
  return matrix->get(l,c);
}

#endif
